/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.List;
import java.util.Map;

import org.xmloperator.lambda3.exception.UndefinedEliminationException;
import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;
import org.xmloperator.lambda3.impl.coding.Combinators;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Application;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.model.Variable;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.util.Stack;

/**
 * Standard Application implementation.
 */
final class ApplicationImpl extends TermImpl implements Application {
	static int creationCount = 0;
	static int breakCount = 0;
	static final Stack<ApplicationImpl> reusableApplicationStack = new Stack<ApplicationImpl>(
			ApplicationImpl.class, REUSE_CAPACITY);

	/**
	 * Constructs and returns a new Application.
	 * 
	 * @param function
	 *            a function Term.
	 * @param argument
	 *            an argument Term.
	 * @return
	 */
	static final ApplicationImpl newApplication(TermImpl function,
			TermImpl argument) {
		if (function == null || argument == null)
			throw new IllegalArgumentException();
		ApplicationImpl application = reusableApplicationStack.pop();
		if (application == null)
			application = new ApplicationImpl();
		application.function = function;
		function.referenceCount++;
		application.argument = argument;
		argument.referenceCount++;
		creationCount++;
		return application;
	}

	TermImpl function;
	TermImpl argument;

	private ApplicationImpl() {
	}

	/**
	 * Replaces the function part of this Application.
	 * 
	 * @param newFunction
	 *            a new function.
	 */
	final void replaceFunction(TermImpl newFunction) {
		if (this.function != null)
			this.function.referenceCount--;
		this.function = newFunction;
		if (this.function != null)
			this.function.referenceCount++;
	}

	/**
	 * Replaces the argument part of this Application.
	 * 
	 * @param newArgument
	 *            a new argument.
	 */
	final void replaceArgument(TermImpl newArgument) {
		if (this.argument != null)
			this.argument.referenceCount--;
		this.argument = newArgument;
		if (this.argument != null)
			this.argument.referenceCount++;
	}

	// ////////////////////
	// Term implementation
	// ////////////////////

	public final boolean isApplication() {
		return true;
	}

	public final boolean isBroken() {
		return this.function == null || this.argument == null;
	}

	public final Term abstractionElimination(TermRecycler termRecycler) {
		TermImpl transformedFunction = (TermImpl) this.function
				.abstractionElimination(termRecycler);
		TermImpl transformedArgument = (TermImpl) this.argument
				.abstractionElimination(termRecycler);
		return newApplication(transformedFunction, transformedArgument);
	}

	// ///////////////////////////
	// Application implementation
	// ///////////////////////////

	public final Term getFunction() {
		return this.function;
	}

	public final Term getArgument() {
		return this.argument;
	}

	public final boolean isRedex() {
		return this.function.isAbstraction()
				|| (this.function.isConstant() && ((Constant) this.function)
						.acceptArgument(this.argument));
	}

	// //////////////////
	// TermImpl override
	// //////////////////

	final void breakIt() {
		if (TermImpl.ASSERTION && this.referenceCount != 0)
			throw new StillReferencedTermException(this);
		if (this.function == null || this.argument == null)
			throw new BrokenTermException(this);
		this.function.referenceCount--;
		this.function = null;
		this.argument.referenceCount--;
		this.argument = null;
		reusableApplicationStack.push(this);
		breakCount++;
	}

	final TermImpl abstractionElimination(Variable boundVariable,
			TermRecycler termRecycler) {
		boolean isBoundVariableUsedInFunction = this.function == boundVariable
				|| AssertUtils.referenceCountOf(this.function, boundVariable) > 0;
		boolean isBoundVariableUsedInArgument = this.argument == boundVariable
				|| AssertUtils.referenceCountOf(this.argument, boundVariable) > 0;
		Term transformedFunction = isBoundVariableUsedInFunction ? this.function
				.abstractionElimination(boundVariable, termRecycler)
				: this.function.abstractionElimination(termRecycler);
		try {
			Term transformedArgument = isBoundVariableUsedInArgument ? this.argument
					.abstractionElimination(boundVariable, termRecycler)
					: this.argument.abstractionElimination(termRecycler);
			if (isBoundVariableUsedInFunction && isBoundVariableUsedInArgument)
				return (ApplicationImpl) Combinators.newS().applyTo(
						transformedFunction).applyTo(transformedArgument);
			if (isBoundVariableUsedInFunction)
				return (ApplicationImpl) Combinators.newC().applyTo(
						transformedFunction).applyTo(transformedArgument);
			if (isBoundVariableUsedInArgument)
				return (ApplicationImpl) Combinators.newB().applyTo(
						transformedFunction).applyTo(transformedArgument);
			throw new IllegalStateException();
		} catch (UndefinedEliminationException e) {
			termRecycler.breakTermDeeply(transformedFunction);
			throw new UndefinedEliminationException();
		}
	}

	final TermType assignType(List<Variable> boundVariableStack,
			List<TermType> boundVariableTypeStack,
			Map<Term, TermType> fromSharedTermToType) {
		if (this.referenceCount > 1) {
			TermType applicationType = fromSharedTermToType.get(this);
			if (applicationType != null)
				return applicationType;
		}
		TermType functionType = this.function.assignType(boundVariableStack,
				boundVariableTypeStack, fromSharedTermToType);
		TermType argumentType = this.argument.assignType(boundVariableStack,
				boundVariableTypeStack, fromSharedTermToType);
		TermType applicationType = TermTypeFactory.newVariableTermType();
		TermTypeFactory.newFunctionTermType(argumentType, applicationType)
				.setEquivalence(functionType);
		if (this.referenceCount > 1)
			fromSharedTermToType.put(this, applicationType);
		return applicationType;
	}
}
