/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.xmloperator.lambda3.exception.AssertionContradictedException;
import org.xmloperator.lambda3.exception.UnimplementedUseCaseException;
import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Utilities on free variables and abstractions.
 */
public abstract class FreeUtils {

	/**
	 * Builds the set of variable that are not bound inside a root Term.
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param freeVariableSet
	 *            the resulting set of free variables.
	 */
	public static final void buildFreeVariableSet(final Term rootTerm,
			final Set<Variable> freeVariableSet) {
		if (rootTerm == null || !freeVariableSet.isEmpty())
			throw new IllegalArgumentException();
		Map<Term, Set<Variable>> fromSharedTermToVariableSet = new HashMap<Term, Set<Variable>>();
		Map<Term, Set<Variable>> fromFunctionToVariableSet = new HashMap<Term, Set<Variable>>();
		List<TermImpl> termStack = new ArrayList<TermImpl>();
		List<Boolean> todoArgumentStack = new ArrayList<Boolean>();
		TermImpl term = (TermImpl) rootTerm;
		boolean todoArgument = false;
		boolean onReturn = false;
		while (true) {
			if (term.isVariable()) {
				if (!freeVariableSet.isEmpty())
					throw new AssertionContradictedException();
				freeVariableSet.add((Variable) term);
				if (termStack.isEmpty())
					return;
				int index = termStack.size() - 1;
				term = termStack.remove(index);
				todoArgument = todoArgumentStack.remove(index).booleanValue();
				onReturn = !todoArgument;
			} else {
				Set<Variable> variableSet = fromSharedTermToVariableSet
						.get(term);
				if (variableSet != null) {
					if (!freeVariableSet.isEmpty())
						throw new AssertionContradictedException();
					freeVariableSet.addAll(variableSet);
					if (termStack.isEmpty())
						throw new AssertionContradictedException();
					int index = termStack.size() - 1;
					term = termStack.remove(index);
					todoArgument = todoArgumentStack.remove(index)
							.booleanValue();
					onReturn = !todoArgument;
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					if (onReturn) {
						freeVariableSet.remove(abstraction.boundVariable);
						if (termStack.isEmpty())
							return;
						if (abstraction.referenceCount > 1)
							fromSharedTermToVariableSet.put(abstraction,
									freeVariableSet);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.booleanValue();
						onReturn = !todoArgument;
					} else {
						if (!freeVariableSet.isEmpty())
							throw new AssertionContradictedException();
						termStack.add(term);
						todoArgumentStack.add(Boolean.FALSE);
						term = abstraction.body;
						if (term == null)
							throw new BrokenTermException(rootTerm, abstraction);
						todoArgument = false;
					}
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (onReturn) {
						Set<Variable> functionVariableSet = fromFunctionToVariableSet
								.remove(application);
						if (functionVariableSet == null)
							throw new AssertionContradictedException();
						freeVariableSet.addAll(functionVariableSet);
						if (termStack.isEmpty())
							return;
						if (application.referenceCount > 1)
							fromSharedTermToVariableSet.put(application,
									freeVariableSet);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.booleanValue();
						onReturn = !todoArgument;
					} else {
						if (todoArgument) {
							Set<Variable> functionVariableSet = new HashSet<Variable>();
							functionVariableSet.addAll(freeVariableSet);
							fromFunctionToVariableSet.put(application,
									functionVariableSet);
							freeVariableSet.clear();
							termStack.add(term);
							todoArgumentStack.add(Boolean.FALSE);
							term = application.argument;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = false;
						} else {
							if (!freeVariableSet.isEmpty())
								throw new AssertionContradictedException();
							termStack.add(term);
							todoArgumentStack.add(Boolean.TRUE);
							term = application.function;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = false;
						}
					}
				} else if (term.isConstant())
					throw new UnimplementedUseCaseException();
				else
					throw new IllegalStateException();
			}
		}
	}

	/**
	 * Builds the set of free abstractions inside a root Term. If the root Term
	 * is an Abstraction then it is excluded.
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param freeAbstractionSet
	 *            the resulting set of free abstractions.
	 */
	public static final void buildFreeAbstractionSet(final Term rootTerm,
			final Set<Abstraction> freeAbstractionSet) {
		if (rootTerm == null || !freeAbstractionSet.isEmpty())
			throw new IllegalArgumentException();
		Set<Variable> rootFreeVariableSet = new HashSet<Variable>();
		buildFreeVariableSet(rootTerm, rootFreeVariableSet);
		Set<Variable> freeVariableSet = new HashSet<Variable>();
		Map<Term, Set<Variable>> fromSharedTermToVariableSet = new HashMap<Term, Set<Variable>>();
		Map<Term, Set<Variable>> fromFunctionToVariableSet = new HashMap<Term, Set<Variable>>();
		List<TermImpl> termStack = new ArrayList<TermImpl>();
		List<Boolean> todoArgumentStack = new ArrayList<Boolean>();
		TermImpl term = (TermImpl) rootTerm;
		boolean todoArgument = false;
		boolean onReturn = false;
		while (true) {
			if (term.isVariable()) {
				if (!freeVariableSet.isEmpty())
					throw new AssertionContradictedException();
				freeVariableSet.add((Variable) term);
				if (termStack.isEmpty())
					return;
				int index = termStack.size() - 1;
				term = termStack.remove(index);
				todoArgument = todoArgumentStack.remove(index).booleanValue();
				onReturn = !todoArgument;
			} else {
				Set<Variable> variableSet = fromSharedTermToVariableSet
						.get(term);
				if (variableSet != null) {
					if (!freeVariableSet.isEmpty())
						throw new AssertionContradictedException();
					freeVariableSet.addAll(variableSet);
					if (termStack.isEmpty())
						throw new AssertionContradictedException();
					int index = termStack.size() - 1;
					term = termStack.remove(index);
					todoArgument = todoArgumentStack.remove(index)
							.booleanValue();
					onReturn = !todoArgument;
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					if (onReturn) {
						freeVariableSet.remove(abstraction.boundVariable);
						if (termStack.isEmpty())
							return;
						if (rootFreeVariableSet.containsAll(freeVariableSet))
							freeAbstractionSet.add(abstraction);
						if (abstraction.referenceCount > 1)
							fromSharedTermToVariableSet.put(abstraction,
									freeVariableSet);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.booleanValue();
						onReturn = !todoArgument;
					} else {
						if (!freeVariableSet.isEmpty())
							throw new AssertionContradictedException();
						termStack.add(term);
						todoArgumentStack.add(Boolean.FALSE);
						term = abstraction.body;
						if (term == null)
							throw new BrokenTermException(rootTerm, abstraction);
						todoArgument = false;
					}
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (onReturn) {
						Set<Variable> functionVariableSet = fromFunctionToVariableSet
								.remove(application);
						if (functionVariableSet == null)
							throw new AssertionContradictedException();
						freeVariableSet.addAll(functionVariableSet);
						if (termStack.isEmpty())
							return;
						if (application.referenceCount > 1)
							fromSharedTermToVariableSet.put(application,
									freeVariableSet);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.booleanValue();
						onReturn = !todoArgument;
					} else {
						if (todoArgument) {
							Set<Variable> functionVariableSet = new HashSet<Variable>();
							functionVariableSet.addAll(freeVariableSet);
							fromFunctionToVariableSet.put(application,
									functionVariableSet);
							freeVariableSet.clear();
							termStack.add(term);
							todoArgumentStack.add(Boolean.FALSE);
							term = application.argument;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = false;
						} else {
							if (!freeVariableSet.isEmpty())
								throw new AssertionContradictedException();
							termStack.add(term);
							todoArgumentStack.add(Boolean.TRUE);
							term = application.function;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = false;
						}
					}
				} else if (term.isConstant())
					throw new UnimplementedUseCaseException();
				else
					throw new IllegalStateException();
			}
		}
	}
}
