/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Implementation of NumberCoding as Scott numerals.
 * 
 * <pre>
 * Term("@0") = "K"
 * Term("@(n+1)") = "\xf.f @n"
 * Term("@+n") = "[K,@n]", including n=0
 * Term("@-n") = "[F,#@]", excepting n=0
 * Term("@-0") = "[K,@0]"
 * FunctionName ::= "isZero" | "succ" | "pred" | "SisZero" | "Ssucc" | "Spred"
 * 					| "fromChurch" | "toChurch"
 * </pre>
 */
public final class ScottNumberCoding implements NumberCoding {
	public static final NumberCoding NUMBER_CODING = new ScottNumberCoding();

	private static final Variable newVariable() {
		return TermFactory.newVariable();
	}

	/**
	 * Returns a new instance of Term(@0) = "\xy.x" = K.
	 * 
	 * @return a new instance of Term(@0) = "\xy.x" = K.
	 */
	public static final Abstraction newZero() {
		return Combinators.newK();
	}

	/**
	 * Returns a new instance of Term(@n)
	 * 
	 * @param number
	 *            a positive or null number
	 * @return a new instance of Term(@n)
	 */
	public static final Abstraction newNumber(int number) {
		if (number < 0)
			throw new IllegalArgumentException();
		Abstraction term = newZero();
		for (int k = 0; k < number; k++) {
			Variable x = newVariable();
			Variable f = newVariable();
			term = f.applyTo(term).abstractOn(f).abstractOn(x)
					.setCombinatorName("@" + (1 + k));
		}
		return term;
	}

	/**
	 * Returns a new instance of signed number.
	 * 
	 * @param number
	 *            a relative number
	 * @return a new instance of signed number.
	 */
	public static final Abstraction newSignedNumber(int number) {
		Term signTerm = Combinators.newBoolean(number >= 0);
		Term absoluteValueTerm = newNumber(number >= 0 ? number : -number);
		return Combinators.newPair(signTerm, absoluteValueTerm)
				.setCombinatorName("@" + (number >= 0 ? "+" : "") + number);
	}

	/**
	 * Returns a new instance of Term(@isZero) = "[K,\y.F]".
	 * 
	 * @return a new instance of Term(@isZero) = "[K,\y.F]".
	 */
	public static final Abstraction newIsZero() {
		Variable y = newVariable();
		return Combinators.newPair(Combinators.newK(),
				Combinators.newF().abstractOn(y)).setCombinatorName("@isZero");
	}

	/**
	 * Returns a new instance of Term(@SisZero) = "\x.@isZero(xF)".
	 * 
	 * @return a new instance of Term(@SisZero) = "\x.@isZero(xF)".
	 */
	public static final Term newSignedIsZero() {
		Variable x = newVariable();
		return newIsZero().applyTo(x.applyTo(Combinators.newF())).abstractOn(x)
				.setCombinatorName("@SisZero");
	}

	/**
	 * Returns a new instance of Term(@+0) = "[K,@0]".
	 * 
	 * @return a new instance of Term(@+0) = "[K,@0]".
	 */
	public static final Abstraction newSignedZero() {
		return Combinators.newPair(Combinators.newK(), newZero())
				.setCombinatorName("@+0");
	}

	/**
	 * Returns a new instance of the function "\x.@SisZero x[K,@0]x", which
	 * normalizes "@-0" in "#+0".
	 * 
	 * @return a new instance of the function "\x.@SisZero x[K,@0]x", which
	 *         normalizes "@-0" in "@+0".
	 */
	private static final Term newNormalizeSign() {
		Variable x = newVariable();
		return newSignedIsZero().applyTo(x).applyTo(newSignedZero()).applyTo(x)
				.abstractOn(x).setCombinatorName("@normSign");
	}

	/**
	 * Returns a new instance of Term(@succ) = "\zxy.yz"
	 * 
	 * @return a new instance of Term(@succ) = "\zxy.yz"
	 */
	public static final Abstraction newSucc() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable z = newVariable();
		return y.applyTo(z).abstractOn(y).abstractOn(x).abstractOn(z)
				.setCombinatorName("@succ");
	}

	/**
	 * Returns a new instance of Term(@pred) = "\z.z@0I = [@0,I]"
	 * 
	 * @return a new instance of Term(@pred) = "\z.z@0I = [@0,I]"
	 */
	public static final Abstraction newPred() {
		return Combinators.newPair(newZero(), Combinators.newI())
				.setCombinatorName("@pred");
	}

	/**
	 * Returns a new instance of Term(@Ssucc)
	 * 
	 * <pre>
	 * "\x.xK[K,@succ (xF)](@Snormalize [F,@pred (xF)])"
	 * </pre>
	 * 
	 * @return a new instance of Term(@Ssucc)
	 */
	public static final Term newSignedSucc() {
		Abstraction K = Combinators.newK();
		Abstraction F = Combinators.newF();
		Variable x = newVariable();
		Term xK = x.applyTo(K);
		Term xF = x.applyTo(F);
		Term caseK = Combinators.newPair(K, newSucc().applyTo(xF));
		Term caseF = newNormalizeSign().applyTo(
				Combinators.newPair(F, newPred().applyTo(xF)));
		return xK.applyTo(caseK).applyTo(caseF).abstractOn(x)
				.setCombinatorName("@Ssucc");
	}

	/**
	 * Returns a new instance of Term(@Spred)
	 * 
	 * <pre>
	 * "\x.xK(@SisZero x[F,@1][K,@pred (xF)])[F,@succ (xF)]"
	 * </pre>
	 * 
	 * @return a new instance of Term(@Spred)
	 */
	public static final Term newSignedPred() {
		Abstraction K = Combinators.newK();
		Abstraction F = Combinators.newF();
		Variable x = newVariable();
		Term xK = x.applyTo(K);
		Term xF = x.applyTo(F);
		Term caseK = newSignedIsZero().applyTo(x).applyTo(
				Combinators.newPair(F, newNumber(1))).applyTo(
				Combinators.newPair(K, newPred().applyTo(xF)));
		Term caseF = Combinators.newPair(F, newSucc().applyTo(xF));
		return xK.applyTo(caseK).applyTo(caseF).abstractOn(x)
				.setCombinatorName("@Spred");
	}

	/**
	 * Returns a new instance of Church to Scott numeral transformer.
	 * 
	 * <pre>
	 * \n.n(\txy.yt)K
	 * </pre>
	 * 
	 * @return a new instance of Church to Scott numeral transformer.
	 */
	public static final Abstraction newFromChurch() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable t = newVariable();
		Variable n = newVariable();
		Abstraction f = y.applyTo(t).abstractOn(y).abstractOn(x).abstractOn(t);
		return n.applyTo(f).applyTo(Combinators.newK()).abstractOn(n)
				.setCombinatorName("@fromChurch");
	}

	/**
	 * Returns a new instance of Scott to Church numeral transformer.
	 * 
	 * <pre>
	 * YG, G = \tn.(@isZero n)F(#succ (t(nII)))
	 * </pre>
	 * 
	 * @return a new instance of Scott to Church numeral transformer.
	 */
	public static final Term newToChurch() {
		Abstraction I = Combinators.newI();
		Variable t = newVariable();
		Variable n = newVariable();
		Term pred_n = n.applyTo(I).applyTo(I);
		Term caseNotZero = ChurchNumberCoding.newSucc().applyTo(
				t.applyTo(pred_n));
		Term G = newIsZero().applyTo(n).applyTo(Combinators.newF()).applyTo(
				caseNotZero).abstractOn(n).abstractOn(t).setCombinatorName(
				"@toChurch");
		return Combinators.newY().applyTo(G);
	}

	// ////////////////////////////
	// NumberCoding implementation
	// ////////////////////////////

	public char getIntroductionChar() {
		return NumberCoding.SCOTT;
	}

	public Term codeNumber(boolean isSigned, int number) {
		return isSigned ? newSignedNumber(number) : newNumber(number);
	}

	public Term codeFunction(String functionName) {
		if (functionName.equals(ISZERO))
			return newIsZero();
		if (functionName.equals(SUCC))
			return newSucc();
		if (functionName.equals(PRED))
			return newPred();
		if (functionName.equals(SIGNED_ISZERO))
			return newSignedIsZero();
		if (functionName.equals(SIGNED_SUCC))
			return newSignedSucc();
		if (functionName.equals(SIGNED_PRED))
			return newSignedPred();
		if (functionName.equals(FROM_CHURCH))
			return newFromChurch();
		if (functionName.equals(TO_CHURCH))
			return newToChurch();
		return null;
	}
}
