/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.model;

import org.xmloperator.lambda3.exception.term.ExternalBoundVariableException;

/**
 * A Term that is an abstraction.
 */
public interface Abstraction extends Term {

	/**
	 * @return the Variable attached to this Abstraction.
	 */
	public Variable getBoundVariable();

	/**
	 * @return the body Term of this Abstraction.
	 */
	public Term getBody();

	/**
	 * Returns <tt>true</tt> if this Abstraction is closed.
	 * 
	 * @return <tt>true</tt> if this Abstraction is closed.
	 */
	public boolean isClosed();

	/**
	 * Returns the combinator name of this Abstraction if any.
	 * 
	 * @return the combinator name of this Abstraction if any.
	 */
	public String getCombinatorName();

	/**
	 * This Abstraction must be closed i.e. not containing any Variable bound to
	 * an external Abstraction. Set its combinator name.
	 * 
	 * @param name
	 *            a combinator name.
	 * @throws ExternalBoundVariableException
	 *             if some Variable contained within this Abstraction is bound
	 *             to an external Abstraction.
	 * @return this Abstraction.
	 */
	public Abstraction setCombinatorName(String name);

	/**
	 * Applies if possible Eta reduction on this Term.
	 * 
	 * <pre>
	 * if M doesn't contain x then
	 * \x.Mx -> M
	 * </pre>
	 */
	public Term etaReduce();

	/**
	 * Returns <tt>true</tt> if this Term is I = \x.x, <tt>false</tt> elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is I = \x.x, <tt>false</tt> elsewhere.
	 */
	public boolean isI();

	/**
	 * Returns <tt>true</tt> if this Term is K = \xy.x, <tt>false</tt>
	 * elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is K = \xy.x, <tt>false</tt>
	 *         elsewhere.
	 */
	public boolean isK();

	/**
	 * Returns <tt>true</tt> if this Term is F = \xy.y, <tt>false</tt>
	 * elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is F = \xy.y, <tt>false</tt>
	 *         elsewhere.
	 */
	public boolean isF();

	/**
	 * Returns <tt>true</tt> if this Term is S = \xyz.xz(yz), <tt>false</tt>
	 * elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is S = \xyz.xz(yz), <tt>false</tt>
	 *         elsewhere.
	 */
	public boolean isS();

	/**
	 * Returns <tt>true</tt> if this Term is B = \xyz.x(yz), <tt>false</tt>
	 * elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is B = \xyz.x(yz), <tt>false</tt>
	 *         elsewhere.
	 */
	public boolean isB();

	/**
	 * Returns <tt>true</tt> if this Term is C = \xyz.xzy, <tt>false</tt>
	 * elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is C = \xyz.xzy, <tt>false</tt>
	 *         elsewhere.
	 */
	public boolean isC();

	/**
	 * Returns <tt>true</tt> if this Term is Y = \x.(\y.x(yy))(\y.x(yy)),
	 * <tt>false</tt> elsewhere.
	 * 
	 * @return <tt>true</tt> if this Term is Y = \x.(\y.x(yy))(\y.x(yy)),
	 *         <tt>false</tt> elsewhere.
	 */
	public boolean isY();

	/**
	 * Returns <tt>true</tt> if this Term is a tuple.
	 * <p>
	 * A Term \x.xABC is a tupple [A,B,C] if neither A, B or C contains x.
	 * </p>
	 * <p>
	 * \x.x is the empty tuple [].
	 * </p>
	 * 
	 * @return <tt>true</tt> if this Term is a tuple.
	 */
	public boolean isTuple();

	/**
	 * Returns the Church numeral that this Abstraction represents or -1 if
	 * none.
	 * 
	 * @return the Church numeral that this Abstraction represents or -1 if
	 *         none.
	 */
	public int getChurchNumeral();
}
