/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.model;

import org.xmloperator.lambda3.model.type.TermType;

/**
 * A term (or a sub_term) of the lambda-calculus.
 * <p>
 * Makes use of term sharing using a Directed Acyclic Graph (DAG). This
 * technique is explained in the following document :
 * </p>
 * <p>
 * http://pauillac.inria.fr/~levy/pubs/88icot.pdf
 * </p>
 * <p>
 * A Term is either a Variable, a Constant, an Abstraction or an Application
 * </p>
 * 
 * <pre>
 * -- Life cycle of a Variable :
 * New or Reuse -> Free (immediately)
 * Free -> Bound
 * Bound -> Broken
 * Broken -> Reusable or Garbage (immediately)
 * Reusable -> Reuse
 * 
 * -- Life cycle of an Abstraction or an Application
 * -- or a constant with more than one argument:
 * New or Reuse -> Usable (immediately)
 * Usable -> Broken
 * Broken -> Reusable or Garbage (immediately)
 * Reusable -> Reuse
 * </pre>
 */
public interface Term {

	/**
	 * Returns <tt>true</tt> if this Term is a Variable.
	 * 
	 * @return <tt>true</tt> if this Term is a Variable.
	 */
	public boolean isVariable();

	/**
	 * Returns <tt>true</tt> if this Term is a Constant.
	 * 
	 * @return <tt>true</tt> if this Term is a Constant.
	 */
	public boolean isConstant();

	/**
	 * Returns <tt>true</tt> if this Term is an Abstraction.
	 * 
	 * @return <tt>true</tt> if this Term is an Abstraction.
	 */
	public boolean isAbstraction();

	/**
	 * Returns <tt>true</tt> if this Term is an Application.
	 * 
	 * @return <tt>true</tt> if this Term is an Application.
	 */
	public boolean isApplication();

	/**
	 * Returns <tt>true</tt> if this Term is broken.
	 * 
	 * @return <tt>true</tt> if this Term is broken.
	 */
	public boolean isBroken();

	/**
	 * Returns the term reference count.
	 * 
	 * @return the term reference count.
	 */
	public int getReferenceCount();

	/**
	 * Builds and returns an Abstraction whose body is this Term. In this body
	 * the given freeVariable (each instance of it) become the bound Variable of
	 * the resulted Abstraction. This constructs a Directed Acyclic Graph (DAG).
	 * 
	 * @param freeVariable
	 *            a FreeVariable possibly referenced by this Term.
	 * @return the built Abstraction.
	 */
	public Abstraction abstractOn(Variable freeVariable);

	/**
	 * Builds and returns an Application whose function is this Term and the
	 * argument is given. This constructs a Directed Acyclic Graph (DAG).
	 * 
	 * @param argument
	 *            a Term.
	 * @return the build Application.
	 */
	public Application applyTo(Term argument);

	/**
	 * Returns a new Term, equivalent to this one but only with combinators.
	 * 
	 * <pre>
	 * T[x] => x
	 * T[(E1 E2)] => (T[E1] T[E2])
	 * T[\x.E] => (K T[E]) (if x is not free in E)
	 * T[\x.x] => I
	 * T[\xy.x] => K
	 * T[\xyz.x(yz)] => B
	 * T[\xyz.xzy] => C
	 * T[\xyz.xz(yz)] => S
	 * T[\xy.E] => T[\x.T[\y.E]] (if x is free in E)
	 * T[\x.(E1 E2)] => (S T[\x.E1] T[\x.E2]) (if x is free in both E1 and E2)
	 * T[\x.(E1 E2)] => (C T[\x.E1] T[E2]) (if x is free in E1 but not E2)
	 * T[\x.(E1 E2)] => (B T[E1] T[\x.E2]) (if x is free in E2 but not E1)
	 * </pre>
	 * 
	 * @param termRecycler
	 *            a TermRecycler.
	 * @return a new Term, equivalent to this one but only with combinators.
	 */
	public Term abstractionElimination(TermRecycler termRecycler);

	/**
	 * Assigns a type to this Term, if possible, and returns it.
	 * <p>
	 * This Term must be closed.
	 * </p>
	 * 
	 * @return an assigned type, if possible.
	 * @throws UntypableTermException
	 *             if it is not possible to assign a type to this Term.
	 * @throws NotClosedTermException
	 *             if this Term is not closed.
	 */
	public TermType assignType();
}
